################ Initialization ###############
#                                             #
#               By: Colin Yancey              #
#             Created October 2025            #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################


from .optimizerResults import (
    CompiledOptimizationData
)


__all__ = [
    "CompiledOptimizationData"
]