########### Iteration Parsing Handler #########
#                                             #
#               By: Colin Yancey              #
#             Created January 2026            #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################


###############################################


# Returns a list of iterations, having a selection frequency based on the size of the intermediate count and the iteration count.
# def getIterationList(intermediateCount = ,)